# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import os
import warnings
from functools import reduce
from test.utils import undeploy_all_version

import joblib
import numpy as np
import pytest
import xgboost as xgb
from merlin.model import ModelType, PyFuncModel, PyFuncV3Model
from merlin.model_schema import ModelSchema
from merlin.observability.inference import InferenceSchema, RegressionOutput, ValueType
from merlin.pyfunc import ModelInput, ModelOutput, Values
from merlin.resource_request import ResourceRequest
from sklearn import svm
from sklearn.datasets import load_iris

import merlin

warnings.filterwarnings("ignore")

request_json = {"instances": [[2.8, 1.0, 6.8, 0.4], [3.1, 1.4, 4.5, 1.6]]}

XGB_PATH = os.path.join("test/pyfunc/", "model_1.bst")
SKLEARN_PATH = os.path.join("test/pyfunc/", "model_2.joblib")


class EnsembleModel(PyFuncModel):
    def initialize(self, artifacts):
        self._model_1 = xgb.Booster(model_file=artifacts["xgb_model"])
        self._model_2 = joblib.load(artifacts["sklearn_model"])

    def infer(self, model_input):
        inputs = np.array(model_input["instances"])
        dmatrix = xgb.DMatrix(model_input["instances"])
        result_1 = self._model_1.predict(dmatrix)
        result_2 = self._model_2.predict_proba(inputs)
        return {"predictions": ((result_1 + result_2) / 2).tolist()}


class EnvVarModel(PyFuncModel):
    def initialize(self, artifacts):
        self.env_var = {}
        self.env_var["workers"] = os.environ.get("WORKERS")
        self.env_var["env_var_1"] = os.environ.get("ENV_VAR_1")
        self.env_var["env_var_2"] = os.environ.get("ENV_VAR_2")

    def infer(self, model_input):
        return self.env_var


class ModelObservabilityModel(PyFuncV3Model):
    def initialize(self, artifacts):
        self._feature_names = [
            "sepal_length",
            "sepal_width",
            "petal_length",
            "petal_width",
        ]
        self._target_names = ["prediction_score"]
        self._model = xgb.Booster(model_file=artifacts["xgb_model"])

    def preprocess(self, request: dict, **kwargs) -> ModelInput:
        features_data = request["instances"]
        return ModelInput(
            prediction_ids=["prediction_1", "prediction_2"],
            features=Values(columns=self._feature_names, data=features_data),
        )

    def infer(self, model_input: ModelInput) -> ModelOutput:
        dmatrix = xgb.DMatrix(model_input.features.data)
        outputs = self._model.predict(dmatrix).tolist()

        def max(first, second):
            if first > second:
                return first
            return second

        prediction_outputs = [[reduce(max, row)] for row in outputs]

        return ModelOutput(
            prediction_ids=model_input.prediction_ids,
            predictions=Values(columns=self._target_names, data=prediction_outputs),
        )

    def postprocess(self, model_output: ModelOutput, request: dict) -> dict:
        return {"predictions": model_output.predictions.data}


def train_xgboost_model(X, y):
    model_1_dir = "test/pyfunc/"
    BST_FILE = "model_1.bst"
    dtrain = xgb.DMatrix(X, label=y)
    param = {
        "max_depth": 6,
        "eta": 0.1,
        "silent": 1,
        "nthread": 4,
        "num_class": 3,
        "objective": "multi:softprob",
    }
    xgb_model = xgb.train(params=param, dtrain=dtrain)
    model_1_path = os.path.join(model_1_dir, BST_FILE)
    xgb_model.save_model(model_1_path)
    return model_1_path


def train_sklearn_model(X, y):
    model_2_dir = "test/pyfunc/"
    MODEL_FILE = "model_2.joblib"
    model_2_path = os.path.join(model_2_dir, MODEL_FILE)

    clf = svm.SVC(gamma="scale", probability=True)
    clf.fit(X, y)
    joblib.dump(clf, model_2_path)
    return model_2_path


@pytest.mark.pyfunc
@pytest.mark.integration
@pytest.mark.dependency()
def test_pyfunc(integration_test_url, project_name, use_google_oauth, requests):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("pyfunc-sample", ModelType.PYFUNC)

    undeploy_all_version()
    with merlin.new_model_version() as v:
        iris = load_iris()
        y = iris["target"]
        X = iris["data"]
        xgb_path = train_xgboost_model(X, y)
        sklearn_path = train_sklearn_model(X, y)
        v.log_pyfunc_model(
            model_instance=EnsembleModel(),
            conda_env="test/pyfunc/env.yaml",
            code_dir=["test"],
            artifacts={"xgb_model": xgb_path, "sklearn_model": sklearn_path},
        )

    endpoint = merlin.deploy(v)

    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])

    merlin.undeploy(v)


@pytest.mark.pyfunc
@pytest.mark.integration
@pytest.mark.dependency()
def test_pyfunc_image_builder_resource_request(
    integration_test_url, project_name, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("pyfunc-image-builder", ModelType.PYFUNC)

    undeploy_all_version()
    with merlin.new_model_version() as v:
        iris = load_iris()
        y = iris["target"]
        X = iris["data"]
        xgb_path = train_xgboost_model(X, y)
        sklearn_path = train_sklearn_model(X, y)
        v.log_pyfunc_model(
            model_instance=EnsembleModel(),
            conda_env="test/pyfunc/env.yaml",
            code_dir=["test"],
            artifacts={"xgb_model": xgb_path, "sklearn_model": sklearn_path},
        )

    image_builder_resource_request = ResourceRequest(
        cpu_request="2", memory_request="4Gi"
    )
    endpoint = merlin.deploy(
        v, image_builder_resource_request=image_builder_resource_request
    )

    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])

    merlin.undeploy(v)


@pytest.mark.pyfunc
@pytest.mark.integration
def test_pyfunc_env_vars(
    integration_test_url, project_name, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("pyfunc-env-vars-sample", ModelType.PYFUNC)

    undeploy_all_version()
    with merlin.new_model_version() as v:
        v.log_pyfunc_model(
            model_instance=EnvVarModel(),
            conda_env="test/pyfunc/env.yaml",
            code_dir=["test"],
            artifacts={},
        )

    env_vars = {"WORKERS": "8", "ENV_VAR_1": "1", "ENV_VAR_2": "2"}
    endpoint = merlin.deploy(v, env_vars=env_vars)
    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert resp.json()["workers"] == "8"
    assert resp.json()["env_var_1"] == "1"
    assert resp.json()["env_var_2"] == "2"
    assert env_vars.items() <= endpoint.env_vars.items()

    merlin.undeploy(v)


@pytest.mark.pyfunc
@pytest.mark.integration
@pytest.mark.dependency()
def test_pyfunc_model_observability(
    integration_test_url, project_name, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("pyfunc-mlobs", ModelType.PYFUNC_V3)

    model_schema = ModelSchema(
        spec=InferenceSchema(
            feature_types={
                "sepal_length": ValueType.FLOAT64,
                "sepal_width": ValueType.FLOAT64,
                "petal_length": ValueType.FLOAT64,
                "petal_width": ValueType.FLOAT64,
            },
            model_prediction_output=RegressionOutput(
                prediction_score_column="prediction_score", actual_score_column="actual"
            ),
        )
    )
    undeploy_all_version()
    with merlin.new_model_version(model_schema=model_schema) as v:
        iris = load_iris()
        y = iris["target"]
        X = iris["data"]

        xgb_path = train_xgboost_model(X, y)
        v.log_pyfunc_model(
            model_instance=ModelObservabilityModel(),
            conda_env="test/pyfunc/env.yaml",
            code_dir=["test"],
            artifacts={"xgb_model": xgb_path},
        )

    endpoint = merlin.deploy(v, enable_model_observability=True)

    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json() is not None
    assert len(resp.json()["predictions"]) == len(request_json["instances"])

    merlin.undeploy(v)


# This implementation of PyFuncModel uses the old infer method (no keyword arguments).
# The keywords arguments for infer() method introduced in Merlin 0.5.2.
class OldInferModel(PyFuncModel):
    def initialize(self, artifacts):
        pass

    def infer(self, model_input):
        return model_input


@pytest.mark.pyfunc
@pytest.mark.integration
def test_pyfunc_old_infer(
    integration_test_url, project_name, use_google_oauth, requests
):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("pyfunc-old-infer-sample", ModelType.PYFUNC)

    undeploy_all_version()
    with merlin.new_model_version() as v:
        v.log_pyfunc_model(
            model_instance=OldInferModel(),
            conda_env="test/pyfunc/env.yaml",
            code_dir=["test"],
            artifacts={},
        )

    endpoint = merlin.deploy(v)
    resp = requests.post(f"{endpoint.url}", json=request_json)

    assert resp.status_code == 200
    assert resp.json()["instances"] == request_json["instances"]

    merlin.undeploy(v)



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from merlin.batch.big_query_util import valid_column, valid_dataset, valid_table_name, valid_table_id, valid_columns


@pytest.mark.parametrize(
    "dataset_name,expected", [
        pytest.param(
            'dataset', True, id="Should valid for dataset_name which contains of letters"
        ),
        pytest.param(
            '_dataset', True, id="Should valid for dataset_name which begin with underscore and followed by letters"
        ),
        pytest.param(
            '1dataset', True, id="Should valid for dataset_name which begin with number and followed by letters"
        ),
        pytest.param(
            '+dataset', False, id="Should not valid for dataset_name which begin with +"
        ),
        pytest.param(
            'dataset?ok', False, id="Should not valid for dataset_name which contains of question mark"
        ),
        pytest.param(
            'BqUQ30d4Ree6HUK4JvgdtP7Im0diryYuTt8AfT0xt1t2oWM6wBDgRiGqpQpiHDV5Tdw4z8mH8MHT1o8OkztgoGQZPShobIrSnSBzLka9fDUnkOv3Fww0kK9zuHoIUlnV1aAvOfF3hZjbnHL4nmUyrhStA2BFGp5kgdD68iMGe9mlC8gB2dOH1Wq3vn4Zjd7ba6kVJeqUWBKic6zltrBCT07odjSYVGKHXEiF8g9tplPqVXd8Nc4POKVq0AwdohOV4T2HnJ5wJog30ZWV44ixVprj1B9di7Tc0AUiIzIetTLrbTijo0ZyUav7CcjXofRRWkOlCseGW3ytbl7RFZFKYutwoqnOKBXP6YImhMfXib44OequZOiFTz0Lx2Tke0rXFQrt3zGyx11Sgq51D22eseQQuR6w6z7xyIggisZ7zGHIBu4hTEig4Qdi4QGyOmtfObHHheZdFOvGjMisE04tqpd9tJ8gJUEkAN10SD58oADc1noJdgun4ShCagbWhoFyVUOMCaVEOTSwKhpBL0TSiHiCrUCathMgP0EnAx3ONTLNKMWaMhOUwkkU2Nc15WizW9jNUnlHcp0s3p9giKXtLJWyyOJK0poKDsewhsfRexcPMjDf5mrdYlDnFHncYjsXc1TZmeAc6GY7ZuAoFg4v5NrQ9FFEuvwYOiy4YBqijfovqGJDtX9pU9C74BRAzPstZcRvsBr1ayvriZbYNZc76Tw8xsgBVaogGm0L9NTdNsyGycS5Q5uerkIsaxJmWw2fXyurt4hXYdsijZN0GF6itzuPzt48iuyiPWMcAI3kguj3gBtKZZ0v1fzhfSe1MAwlvludHPE0fyQYXsVUY4JxmOjXFdQ5lZHltjs7JzjEa9wD3bHlJEQZFV0hyf5cNefaShwAqe4E4cc3tQ720A9MkC9mC6EichP3aRcd64I0GIUjuekUUvkS8v6YPDPitKyHEEYiqJNK1J0k2ZwPNqXPgtTQTZmvo3GBWmgNeovFx9GwgaQjdfx0cLSdQr46pZRLh',
            False,
            id="Should not valid for dataset_name which length more than 1024"
        )
    ]
)
def test_valid_dataset(dataset_name, expected):
    valid = valid_dataset(dataset_name)
    assert expected == valid


@pytest.mark.parametrize(
    "column_name,expected", [
        pytest.param(
            'column', True, id="Should valid if column_name contains all letters"
        ),
        pytest.param(
            '_column', True, id="Should valid if column_name begins with underscore and followed with letters"
        ),
        pytest.param(
            'Column', True, id="Should valid if column_name contains all letters regardless uppercase or lowercase"
        ),
        pytest.param(
            '1column', False, id="Should not valid if column_name begins with number"
        ),
        pytest.param(
            'column?1', False, id="Should not valid if column_name contains ? or any special characters"
        ),
        pytest.param(
            'column?', False, id="Should not valid if column_name contains ? or any special characters(2)"
        ),
        pytest.param(
            'YKbbvPahCfJQAFw7OQmx2LDfrUOoHcy5jWmJTWomeR3jyy2KdWsCCxUDo6heuOQdQtmHdL4fGt6Bjx8vdz2yZ8WCSLBEVK4jp89HZYhBcogbP2FLoS9ePdVlvWGv7ljJ2',
            False,
            id="Should not valid if column_name length more than 128"
        ),
        pytest.param(
            '_TABLE_1', False, id="Should not valid if column_name has prefix _TABLE"
        ),
        pytest.param(
            '_FILE_abc', False, id="Should not valid if column_name has prefix _FILE_"
        ),
        pytest.param(
            '_PARTITION2', False, id="Should not valid if column_name has prefix _PARTITION"
        ),
    ]
)
def test_valid_column(column_name, expected):
    valid = valid_column(column_name)
    assert expected == valid


@pytest.mark.parametrize(
    "table_name,expected", [
        pytest.param(
            'table', True, id="Should valid if table_name is contains all letters"
        ),
        pytest.param(
            '_table', True, id="Should valid if table_name is begin with underscore and followed by letters"
        ),
        pytest.param(
            '1table', True, id="Should valid if table_name is begin with number and followed by letters"
        ),
        pytest.param(
            '+table', False, id="Should not valid if table_name contains of + or any special characters"
        ),
        pytest.param(
            'table?ok', False, id="Should not valid if table_name contains of ? or any special characters"
        ),
        pytest.param(
            'BqUQ30d4Ree6HUK4JvgdtP7Im0diryYuTt8AfT0xt1t2oWM6wBDgRiGqpQpiHDV5Tdw4z8mH8MHT1o8OkztgoGQZPShobIrSnSBzLka9fDUnkOv3Fww0kK9zuHoIUlnV1aAvOfF3hZjbnHL4nmUyrhStA2BFGp5kgdD68iMGe9mlC8gB2dOH1Wq3vn4Zjd7ba6kVJeqUWBKic6zltrBCT07odjSYVGKHXEiF8g9tplPqVXd8Nc4POKVq0AwdohOV4T2HnJ5wJog30ZWV44ixVprj1B9di7Tc0AUiIzIetTLrbTijo0ZyUav7CcjXofRRWkOlCseGW3ytbl7RFZFKYutwoqnOKBXP6YImhMfXib44OequZOiFTz0Lx2Tke0rXFQrt3zGyx11Sgq51D22eseQQuR6w6z7xyIggisZ7zGHIBu4hTEig4Qdi4QGyOmtfObHHheZdFOvGjMisE04tqpd9tJ8gJUEkAN10SD58oADc1noJdgun4ShCagbWhoFyVUOMCaVEOTSwKhpBL0TSiHiCrUCathMgP0EnAx3ONTLNKMWaMhOUwkkU2Nc15WizW9jNUnlHcp0s3p9giKXtLJWyyOJK0poKDsewhsfRexcPMjDf5mrdYlDnFHncYjsXc1TZmeAc6GY7ZuAoFg4v5NrQ9FFEuvwYOiy4YBqijfovqGJDtX9pU9C74BRAzPstZcRvsBr1ayvriZbYNZc76Tw8xsgBVaogGm0L9NTdNsyGycS5Q5uerkIsaxJmWw2fXyurt4hXYdsijZN0GF6itzuPzt48iuyiPWMcAI3kguj3gBtKZZ0v1fzhfSe1MAwlvludHPE0fyQYXsVUY4JxmOjXFdQ5lZHltjs7JzjEa9wD3bHlJEQZFV0hyf5cNefaShwAqe4E4cc3tQ720A9MkC9mC6EichP3aRcd64I0GIUjuekUUvkS8v6YPDPitKyHEEYiqJNK1J0k2ZwPNqXPgtTQTZmvo3GBWmgNeovFx9GwgaQjdfx0cLSdQr46pZRLh',
            False,
            id="Should not valid if table_name length is more than 1024"
        )
    ]
)
def test_valid_table(table_name, expected):
    valid = valid_table_name(table_name)
    assert expected == valid


@pytest.mark.parametrize(
    "source_table,expected", [
        pytest.param(
            'project.dataset.table', True,
            id="Should valid if source_table contains project_id dataset and table and all components are valid"
        ),
        pytest.param(
            'project-with-dash-and-number-007.dataset.table', True,
            id="Should valid if source_table contains project_id which contain number and dash"
        ),
        pytest.param(
            'project. dataset.table', False, id="Should not valid if source_table contains invalid dataset"
        ),
        pytest.param(
            'project.dataset .table', False, id="Should not valid if source_table contains invalid dataset(2)"
        ),
        pytest.param(
            'project.dataset?.table', False, id="Should not valid if source_table contains invalid dataset(3)"
        ),
        pytest.param(
            'project.dataset. table', False, id="Should not valid if source_table contains invalid table"
        ),
        pytest.param(
            'project.dataset.table ', False, id="Should not valid if source_table contains invalid table(2)"
        ),
        pytest.param(
            'project.dataset.table?table', False, id="Should not valid if source_table contains invalid table(3)"
        ),
        pytest.param(
            '.dataset.table', False, id="Should not valid if source_table doesnt have project_id info"
        ),
        pytest.param(
            '    abc.dataset.table', False, id="Should not valid if source_table contains invalid project_id"
        )
    ]
)
def test_valid_source_table(source_table, expected):
    valid = valid_table_id(source_table)
    assert expected == valid


@pytest.mark.parametrize(
    "columns,expected", [
        pytest.param(
            ['column1', 'column2'], True, id="Should valid if all columns contains of letters"
        ),
        pytest.param(
            ['coLumn1', 'column2'], True, id="Should valid if all columns contains of letters(2)"
        ),
        pytest.param(
            ['?column1', 'column2'], False, id="Should valid if one of the columns contains invalid column"
        ),
        pytest.param(
            ['column1', '12column2'], False, id="Should valid if one of the columns contains invalid column"
        ),
    ]
)
def test_valid_columns(columns, expected):
    valid = valid_columns(columns)
    assert expected == valid



# coding: utf-8

"""
    Merlin

    API Guide for accessing Merlin's model management, deployment, and serving functionalities

    The version of the OpenAPI document: 0.14.0
    Generated by OpenAPI Generator (https://openapi-generator.tech)

    Do not edit the class manually.
"""  # noqa: E501


from __future__ import annotations
import json
import pprint
import re  # noqa: F401
from enum import Enum



try:
    from typing import Self
except ImportError:
    from typing_extensions import Self


class FileFormat(str, Enum):
    """
    FileFormat
    """

    """
    allowed enum values
    """
    INVALID_FILE_FORMAT = 'INVALID_FILE_FORMAT'
    CSV = 'CSV'
    PARQUET = 'PARQUET'
    AVRO = 'AVRO'
    JSON = 'JSON'

    @classmethod
    def from_json(cls, json_str: str) -> Self:
        """Create an instance of FileFormat from a JSON string"""
        return cls(json.loads(json_str))





from enum import Enum


class MetricsType(Enum):
    """
    Metrics type to be used for AutoscalingPolicy.

    CONCURRENCY: number of concurrent request handled.
    CPU_UTILIZATION: percentage of CPU utilization.
    MEMORY_UTILIZATION: percentage of Memory utilization.
    RPS: throughput in request per second.
    """

    CONCURRENCY = "concurrency"
    CPU_UTILIZATION = "cpu_utilization"
    MEMORY_UTILIZATION = "memory_utilization"
    RPS = "rps"


class AutoscalingPolicy:
    """
    Autoscaling policy to be used for a deployment.

    """

    def __init__(self, metrics_type: MetricsType, target_value: float):
        self._metrics_type = metrics_type
        self._target_value = target_value

    @property
    def metrics_type(self) -> MetricsType:
        """
        Metrics type to be used for the autoscaling
        :return: MetricsType
        """
        return self._metrics_type

    @property
    def target_value(self) -> float:
        """
        Target metrics value when autoscaling should be performed
        :return: target value
        """
        return self._target_value


RAW_DEPLOYMENT_DEFAULT_AUTOSCALING_POLICY = AutoscalingPolicy(
    MetricsType.CPU_UTILIZATION, 50
)
SERVERLESS_DEFAULT_AUTOSCALING_POLICY = AutoscalingPolicy(MetricsType.CONCURRENCY, 1)



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from os import listdir
from os.path import isdir, isfile, join


def validate_model_dir(model_type, model_dir):
    """
    Validates user-provided model directory based on file structure.
    For tensorflow models, checking is only done on the subdirectory
    with the largest version number.

    :param model_type: type of given model
    :param model_dir: directory containing serialised model file
    """
    from merlin.model import ModelType

    if not isdir(model_dir):
        raise ValueError(f"{model_dir} is not a directory")

    if (
        model_type == ModelType.PYFUNC
        or model_type == ModelType.PYFUNC_V2
        or model_type == ModelType.PYFUNC_V3
        or model_type == ModelType.CUSTOM
    ):
        return

    if model_type == ModelType.TENSORFLOW:
        # for tensorflow model it must have following structure
        # <version>
        # |- variables (directory)
        # |- saved_model.pb
        dirs = listdir(model_dir)
        if len(dirs) != 1:
            raise ValueError(f"{model_dir} must contain a version directory")

        version_dir = join(model_dir, dirs[0])
        variable_dir = join(version_dir, "variables")
        saved_model_path = join(version_dir, "saved_model.pb")
        if not isdir(variable_dir):
            raise ValueError(f"{variable_dir} is not found")

        if not isfile(saved_model_path):
            raise ValueError(f"{saved_model_path} is not found")

        return

    if model_type == ModelType.PYTORCH:
        # for Pytorch model it must have following structure
        # |- config
        # |   |- config.properties
        # |- model-store
        #     |- *.mar
        config_path = join(model_dir, "config", "config.properties")
        if not isfile(config_path):
            raise ValueError(f"{config_path} is not found")

        model_store_dir = join(model_dir, "model-store")
        if not any(fname.endswith(".mar") for fname in listdir(model_store_dir)):
            raise ValueError(f".mar file is not found in {model_store_dir}")

        return

    model_file_map = {
        ModelType.XGBOOST: ["model.bst"],
        ModelType.SKLEARN: ["model.joblib"],
        ModelType.ONNX: ["model.onnx"],
    }
    files = listdir(model_dir)
    if not all([file in files for file in model_file_map[model_type]]):
        raise ValueError(f"{model_file_map[model_type]} is not found in {model_dir}")



from test.batch_integration_test import (
    MODEL_PATH,
    MODEL_PATH_ARTIFACT_KEY,
    IrisClassifier,
)
from test.pyfunc_integration_test import EnvVarModel

import joblib
import pytest
from sklearn import svm
from sklearn.datasets import load_iris

import merlin
from merlin.model import ModelType


@pytest.mark.build_image
@pytest.mark.integration
@pytest.mark.dependency()
def test_build_image_pyfunc(integration_test_url, project_name, use_google_oauth):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("build-image", ModelType.PYFUNC)

    with merlin.new_model_version() as v:
        v.log_pyfunc_model(
            model_instance=EnvVarModel(),
            conda_env="test/pyfunc/env.yaml",
            code_dir=["test"],
            artifacts={},
        )

    image = merlin.build_image(v)
    assert image.exists == True


@pytest.mark.build_image
@pytest.mark.integration
@pytest.mark.dependency()
def test_build_image_batch(integration_test_url, project_name, use_google_oauth):
    merlin.set_url(integration_test_url, use_google_oauth=use_google_oauth)
    merlin.set_project(project_name)
    merlin.set_model("build-image-batch", ModelType.PYFUNC_V2)

    clf = svm.SVC(gamma="scale")
    iris = load_iris()
    X, y = iris.data, iris.target
    clf.fit(X, y)
    joblib.dump(clf, MODEL_PATH)

    with merlin.new_model_version() as v:
        v.log_pyfunc_model(
            model_instance=IrisClassifier(),
            conda_env="test/batch/model/env.yaml",
            code_dir=["test"],
            artifacts={MODEL_PATH_ARTIFACT_KEY: MODEL_PATH},
        )

    image = merlin.build_image(v)
    assert image.exists == True



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import pytest

from merlin.resource_request import ResourceRequest


@pytest.mark.unit
def test_resource_request():
    # Valid resource_request
    ResourceRequest(
        min_replica=1, max_replica=2, cpu_request="100m", memory_request="128Mi"
    )

    # Valid resource_request even though min_replica and max_replica are None
    ResourceRequest(cpu_request="100m", memory_request="128Mi")

    with pytest.raises(
        TypeError,
        match="'>' not supported between instances of 'int' and 'NoneType'",
    ):
        ResourceRequest(min_replica=1, cpu_request="100m", memory_request="128Mi")

    with pytest.raises(
        TypeError,
        match="'>' not supported between instances of 'NoneType' and 'int'",
    ):
        ResourceRequest(max_replica=1, cpu_request="100m", memory_request="128Mi")

    with pytest.raises(
        Exception, match="Min replica must be less or equal to max replica"
    ):
        ResourceRequest(
            min_replica=100, max_replica=2, cpu_request="100m", memory_request="128Mi"
        )

    with pytest.raises(Exception, match="Max replica must be greater than 0"):
        ResourceRequest(
            min_replica=0, max_replica=0, cpu_request="100m", memory_request="128Mi"
        )



from docutils import nodes
from ipypublish.sphinx.notebook.nodes import AdmonitionNode
from ipypublish.sphinx.notebook.nodes import CodeAreaNode
from ipypublish.sphinx.notebook.nodes import FancyOutputNode

def visit_AdmonitionNode(self, node):
    if 'warning' in node['classes']:
        self.visit_warning(node)
    else:
        self.visit_note(node)

def depart_AdmonitionNode(self, node):
    if 'warning' in node['classes']:
        self.depart_warning(node)
    else:
        self.depart_note(node)

def visit_CodeAreaNode(self, node):
    pass

def depart_CodeAreaNode(self, node):
    pass

def visit_FancyOutputNode(self, node):
    raise nodes.SkipNode

def setup(app):
    app.require_sphinx('1.0')

    app.add_node(AdmonitionNode, confluence=
        (visit_AdmonitionNode, depart_AdmonitionNode), override=True)
    app.add_node(CodeAreaNode, confluence=
        (visit_CodeAreaNode, depart_CodeAreaNode), override=True)
    app.add_node(FancyOutputNode, confluence=
        (visit_FancyOutputNode, None), override=True)


# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from enum import Enum
from typing import Dict, Optional

import client
from merlin.autoscaling import (RAW_DEPLOYMENT_DEFAULT_AUTOSCALING_POLICY,
                                SERVERLESS_DEFAULT_AUTOSCALING_POLICY,
                                AutoscalingPolicy, MetricsType)
from merlin.deployment_mode import DeploymentMode
from merlin.environment import Environment
from merlin.logger import Logger
from merlin.protocol import Protocol
from merlin.util import autostr, get_url
from merlin.resource_request import ResourceRequest
from merlin.transformer import Transformer, TransformerType
from merlin.util import extract_optional_value_with_default

class Status(Enum):
    PENDING = 'pending'
    RUNNING = 'running'
    SERVING = 'serving'
    FAILED = 'failed'
    TERMINATED = 'terminated'


@autostr
class VersionEndpoint:
    def __init__(self, endpoint: client.VersionEndpoint, log_url: str = None):
        self._protocol = Protocol.HTTP_JSON
        if endpoint.protocol:
            self._protocol = Protocol(endpoint.protocol)

        self._url = endpoint.url
        if self._protocol == Protocol.HTTP_JSON and endpoint.url is not None and ":predict" not in endpoint.url:
            self._url = f"{endpoint.url}:predict"

        self._status = Status(endpoint.status)
        self._id = endpoint.id
        self._environment_name = endpoint.environment_name
        self._environment = Environment(endpoint.environment) if endpoint.environment is not None else None
        self._env_vars = endpoint.env_vars
        self._logger = Logger.from_logger_response(endpoint.logger)
        self._resource_request = ResourceRequest.from_response(endpoint.resource_request) if endpoint.resource_request is not None else None
        self._deployment_mode = DeploymentMode.SERVERLESS if not endpoint.deployment_mode \
            else DeploymentMode(endpoint.deployment_mode)

        if endpoint.autoscaling_policy is None:
            if self._deployment_mode == DeploymentMode.SERVERLESS:
                self._autoscaling_policy = SERVERLESS_DEFAULT_AUTOSCALING_POLICY
            else:
                self._autoscaling_policy = RAW_DEPLOYMENT_DEFAULT_AUTOSCALING_POLICY
        else:
            self._autoscaling_policy = AutoscalingPolicy(metrics_type=MetricsType(endpoint.autoscaling_policy.metrics_type),
                                                         target_value=endpoint.autoscaling_policy.target_value)

        transformer = endpoint.transformer
        if transformer is not None:
            image = transformer.image if transformer.image is not None else ""
            transformer_type_value = extract_optional_value_with_default(transformer.transformer_type, TransformerType.STANDARD_TRANSFORMER.value)
            transformer_type = TransformerType(transformer_type_value)
            
            transformer_request = None
            if transformer.resource_request is not None:
                transformer_request = ResourceRequest(
                    min_replica=transformer.resource_request.min_replica,
                    max_replica=transformer.resource_request.max_replica,
                    cpu_request=transformer.resource_request.cpu_request,
                    memory_request=transformer.resource_request.memory_request
                )

            env_vars: Dict[str, str] = {}
            self._transformer: Optional[Transformer] = None
            if transformer.env_vars is not None:
                for env_var in transformer.env_vars:
                    if env_var.name is not None and env_var.value is not None:
                        env_vars[env_var.name] = env_var.value

            self._transformer = Transformer(
                image, 
                id=extract_optional_value_with_default(transformer.id, ""),
                enabled=extract_optional_value_with_default(transformer.enabled, False), 
                command=transformer.command,
                args=transformer.args, 
                transformer_type=transformer_type,
                resource_request=transformer_request, 
                env_vars=env_vars,
            )

        if log_url is not None:
            self._log_url = log_url

        self._enable_model_observability = extract_optional_value_with_default(endpoint.enable_model_observability, False)

    @property
    def url(self):
        return self._url

    @property
    def status(self) -> Status:
        return self._status

    @property
    def id(self) -> Optional[str]:
        return self._id

    @property
    def environment_name(self) -> Optional[str]:
        return self._environment_name

    @property
    def environment(self) -> Optional[Environment]:
        return self._environment

    @property
    def env_vars(self) -> Dict[str, str]:
        env_vars = {}
        if self._env_vars:
            for ev in self._env_vars:
                if ev.name is not None and ev.value is not None:
                    env_vars[ev.name] = ev.value
        return env_vars

    @property
    def logger(self) -> Logger:
        return self._logger

    @property
    def log_url(self) -> str:
        return self._log_url

    @property
    def deployment_mode(self) -> DeploymentMode:
        return self._deployment_mode

    @property
    def autoscaling_policy(self) -> AutoscalingPolicy:
        return self._autoscaling_policy

    @property
    def protocol(self) -> Protocol:
        return self._protocol
    
    @property
    def resource_request(self) -> Optional[ResourceRequest]:
        return self._resource_request
    
    @property
    def transformer(self) -> Optional[Transformer]:
        return self._transformer
    
    @property
    def enable_model_observability(self) -> bool:
        return self._enable_model_observability

    def _repr_html_(self):
        return f"""<a href="{self._url}">{self._url}</a>"""


@autostr
class ModelEndpoint:
    def __init__(self, endpoint: client.ModelEndpoint):
        self._protocol = Protocol.HTTP_JSON
        if endpoint.protocol is not None:
            self._protocol = Protocol(endpoint.protocol)

        if self._protocol == Protocol.HTTP_JSON:
            self._url = get_url(f"{endpoint.url}/v1/predict")
        else:
            self._url = endpoint.url
        self._status = Status(endpoint.status)
        self._id = extract_optional_value_with_default(endpoint.id, 0)
        self._environment_name = endpoint.environment_name
        self._environment = Environment(endpoint.environment) if endpoint.environment is not None else None


    @property
    def url(self):
        return self._url

    @property
    def status(self) -> Status:
        return self._status

    @property
    def id(self) -> int:
        return self._id

    @property
    def environment_name(self) -> Optional[str]:
        return self._environment_name

    @property
    def environment(self) -> Optional[Environment]:
        return self._environment

    @property
    def protocol(self) -> Protocol:
        return self._protocol

    def _repr_html_(self):
        return f"""<a href="{self._url}">{self._url}</a>"""



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from abc import ABC, abstractmethod
from typing import Iterable, MutableMapping, Mapping, Any, Optional
from merlin.batch.big_query_util import valid_table_id, valid_columns
import client

class Source(ABC):
    @abstractmethod
    def to_dict(self) -> Mapping[str, Any]:
        pass


class BigQuerySource(Source):
    """
        Source contract for BigQuery to create prediction job
    """

    def __init__(self, table: str, features: Iterable[str],
                 options: MutableMapping[str, str] = None):
        """

        :param table: table id if the source in format of `gcp-project.dataset.table_name`
        :param features: list of features to be used for prediction, it has to match the column name in the source table.
        :param options: additional option to configure source.
        """
        self._table = table
        self._features = features
        self._options = options

    @property
    def table(self) -> str:
        return self._table

    @table.setter
    def table(self, table):
        self._table = table

    @property
    def features(self) -> Iterable[str]:
        return self._features

    @features.setter
    def features(self, features):
        self._features = features

    @property
    def options(self) -> Optional[MutableMapping[str, str]]:
        return self._options

    @options.setter
    def options(self, options):
        self._options = options

    def _validate(self):
        if not self._valid_types():
            raise ValueError("invalid input type")
        if not valid_table_id(self.table):
            raise ValueError(f"invalid table: {self.table}")
        if not valid_columns(self.features):
            raise ValueError(f"invalid features column: {self.features}")

    def _valid_types(self) -> bool:
        if not isinstance(self._table, str):
            return False
        if not isinstance(self._features, list):
            return False
        if self._options is not None and not isinstance(self._options,
                                                        MutableMapping):
            return False

        for feature in self._features:
            if not isinstance(feature, str):
                return False

        return True

    def to_dict(self) -> Mapping[str, Any]:
        self._validate()

        opts = self.options
        if opts is None:
            opts = {}
        return {
            'table': self._table,
            'features': self._features,
            'options': opts
        }

    def to_client_bq_source(self) -> client.PredictionJobConfigBigquerySource:
        opts = {}
        if self.options is not None:
            for k, v in self.options.items():
                opts[k] = v

        return client.PredictionJobConfigBigquerySource(
            table=self._table,
            features=list(self._features),
            options=opts
        )



# Copyright 2020 The Merlin Authors
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.

import client
from typing import Optional

from merlin.resource_request import ResourceRequest

class Environment:
    def __init__(self, env: client.Environment):
        self._name = env.name
        self._cluster = env.cluster
        self._is_default = env.is_default if env.is_default is not None else False
        self._default_resource_request = None
        if env.default_resource_request is not None:
            self._default_resource_request = ResourceRequest(env.default_resource_request.min_replica,
                                                            env.default_resource_request.max_replica,
                                                            env.default_resource_request.cpu_request,
                                                            env.default_resource_request.memory_request)

    @property
    def name(self) -> str:
        return self._name

    @property
    def cluster(self) -> str:
        return self._cluster

    @property
    def is_default(self) -> bool:
        return self._is_default

    @property
    def default_resource_request(self) -> Optional[ResourceRequest]:
        return self._default_resource_request
